#pragma once

#include <thread>
#include <condition_variable>
#include <mutex>
#include <future>
#include <queue>
#include <memory>

class ThreadPool {

    struct Dispatcher {
        virtual void run() = 0;
        virtual ~Dispatcher() = default;
    };

    template< typename Callback, typename... Args >
    struct RealDispatcher : Dispatcher {
        using ReturnType = typename std::result_of< Callback( Args... ) >::type;
        using Tuple = std::tuple< Args... >;

        template< typename... PArgs >
        RealDispatcher( std::promise< ReturnType > promise, Callback &&c, PArgs &&... args ) :
            _promise( std::move( promise ) ),
            _c( std::move( c ) ),
            _storage( std::forward< Args >( args )... )
        {}

        void run() override {

            try {
                call< ReturnType >( 0 );
            }
            catch ( ... ) {
                _promise.set_exception( std::current_exception() );
            }
        }
    private:

        template< typename RT >
        void call( long ) {
            _promise.set_value( callImpl( std::index_sequence_for< Args... >() ) );
        }

        template< typename RT >
        auto call( int )
            -> typename std::enable_if< std::is_same< RT, void >::value >::type
        {
            callImpl( std::index_sequence_for< Args... >() );
            _promise.set_value();
        }

        template< std::size_t... I >
        ReturnType callImpl( std::index_sequence< I... > ) {
            return _c( std::get< I >( _storage )... );
        }

        std::promise< ReturnType > _promise;
        Callback _c;
        Tuple _storage;
    };

    std::queue< std::unique_ptr< Dispatcher > > _queue;
    std::vector< std::thread > _threads;
    std::mutex _mutex;
    std::condition_variable _cv;
    bool _finish;

    void dispatcherManager() {
        while ( true ) {
            std::unique_ptr< Dispatcher > d;
            std::unique_lock< std::mutex > l( _mutex );

            if ( _finish )
                return;
            if ( _queue.empty() ) {
                _cv.wait( l, [this] {
                    return this->_finish || !this->_queue.empty();
                } );
                if ( _finish )
                    return;
            }
            std::swap( d, _queue.front() );

            _queue.pop();
            l.unlock();

            d->run();
        }
    }

    void finish() {
        std::lock_guard< std::mutex > l( _mutex );
        _finish = true;
    }

public:

    struct Exception : std::logic_error {
        Exception() : logic_error( "ThreadPool" ) {}
    };


    ThreadPool( unsigned threads ) :
        _finish( false ),
        _threads( threads )
    {
        if ( threads == 0 )
            throw Exception();

        for ( auto &t : _threads ) {
            t = std::thread( [this] {
                this->dispatcherManager();
            } );
        }
    }
    ThreadPool( const ThreadPool & ) = delete;
    ThreadPool( ThreadPool && ) = delete;
    ThreadPool &operator=( ThreadPool ) = delete;

    ~ThreadPool() {
        finish();
        _cv.notify_all();
        for ( auto &t : _threads )
            t.join();
    }

    template< typename Callback, typename... Args >
    auto enqueue( Callback &&c, Args &&... args )
        -> std::future< typename std::result_of< std::decay_t< Callback >( std::decay_t< Args >... ) >::type >
    {
        using ReturnType = typename std::result_of< std::decay_t< Callback >( std::decay_t< Args >... ) >::type;
        std::promise<ReturnType> promise;
        std::future< ReturnType > future = promise.get_future();
        {
            std::lock_guard< std::mutex > l( _mutex );
            _queue.emplace( new RealDispatcher< std::decay_t< Callback >, std::decay_t< Args >... >(
                std::move( promise ),
                std::move( c ),
                std::forward< Args >( args )... ) );
        } // unlock _mutex
        _cv.notify_one();
        return future;
    }

};
